@@ -6,7 +6,6 @@ import json
 import logging
 import os
 import sys
-from time import time
 
 from tornado import gen
 from tornado.httpclient import AsyncHTTPClient, HTTPError
@@ -15,6 +14,7 @@ from tornado.ioloop import IOLoop
 
 from distributed.compatibility import ConnectionRefusedError
 from distributed.core import read, connect, write
+from distributed.metrics import time
 from distributed.protocol.pickle import dumps
 from distributed.diagnostics.eventstream import eventstream
 from distributed.diagnostics.progress_stream import (progress_stream,
@@ -116,15 +116,11 @@ def task_events(interval, deque, times, index, rectangles, workers, last_seen):
                 if 'compute_start' in msg:
                     deque.append(msg)
                     times.append(msg['compute_start'])
-                    index.append(i)
-                    i += 1
-                    if msg.get('transfer_start') is not None:
+                    count = task_stream_append(rectangles, msg, workers)
+                    for _ in range(count):
                         index.append(i)
                         i += 1
-                    if msg.get('disk_load_start') is not None:
-                        index.append(i)
-                        i += 1
-                    task_stream_append(rectangles, msg, workers)
+
     except StreamClosedError:
         pass  # don't log StreamClosedErrors
     except Exception as e:
