@@ -1,7 +1,7 @@
 from __future__ import print_function, division, absolute_import
 
 from bisect import bisect
-from toolz import valmap
+from operator import add
 
 from bokeh.layouts import row, column
 from bokeh.models import (
@@ -14,6 +14,7 @@ from bokeh.models import (
 from bokeh.models.widgets import DataTable, TableColumn, NumberFormatter
 from bokeh.palettes import Spectral9
 from bokeh.plotting import figure
+from toolz import valmap
 
 from distributed.diagnostics.progress_stream import progress_quads, nbytes_bar
 from distributed.utils import log_errors
@@ -52,6 +53,7 @@ class TaskStream(DashboardComponent):
         """
         self.n_rectangles = n_rectangles
         self.clear_interval = clear_interval
+        self.last = 0
 
         self.source = ColumnDataSource(data=dict(
             start=[], duration=[], key=[], name=[], color=[],
@@ -96,6 +98,7 @@ class TaskStream(DashboardComponent):
         self.root.add_tools(
             hover,
             # export,
+            BoxZoomTool(),
             ResetTool(reset_size=False),
             PanTool(dimensions="width"),
             WheelZoomTool(dimensions="width")
@@ -119,10 +122,12 @@ class TaskStream(DashboardComponent):
 
             # If there has been a significant delay then clear old rectangles
             if rectangles['start']:
-                last_end = old['start'][ind - 1] + old['duration'][ind - 1]
-                if min(rectangles['start']) > last_end + self.clear_interval:
-                    self.source.data.update(rectangles)
-                    return
+                m = min(map(add, rectangles['start'], rectangles['duration']))
+                if m > self.last:
+                    self.last, last = m, self.last
+                    if m > last + self.clear_interval:
+                        self.source.data.update(rectangles)
+                        return
 
             self.source.stream(rectangles, self.n_rectangles)
 
@@ -207,7 +212,7 @@ class TaskProgress(DashboardComponent):
             if messages['tasks']['deque']:
                 self.root.title.text = ("Progress -- total: %(total)s, "
                     "in-memory: %(in-memory)s, processing: %(processing)s, "
-                    "ready: %(ready)s, waiting: %(waiting)s, failed: %(failed)s"
+                    "waiting: %(waiting)s, failed: %(failed)s"
                     % messages['tasks']['deque'][-1])
 
 
@@ -457,15 +462,14 @@ class WorkerTable(DashboardComponent):
             self.source.data.update(data)
 
 
-class ProcessingStacks(DashboardComponent):
-    """ Processing and Stacks distribution per core
+class Processing(DashboardComponent):
+    """ Processing and distribution per core
 
     This shows how many tasks are actively running on each worker and how many
     tasks are enqueued for each worker and how many are in the common pool
     """
     def __init__(self, **kwargs):
-        data = self.processing_update({'processing': {}, 'stacks': {},
-                                  'ncores': {}, 'ready': 0})
+        data = self.processing_update({'processing': {}, 'ncores': {}})
         self.source = ColumnDataSource(data)
 
         x_range = Range1d(-1, 1)
@@ -474,8 +478,6 @@ class ProcessingStacks(DashboardComponent):
              x_range=x_range, id='bk-processing-stacks-plot', **kwargs)
         fig.quad(source=self.source, left=0, right='right', color=Spectral9[0],
                  top='top', bottom='bottom')
-        fig.quad(source=self.source, left='left', right=0, color=Spectral9[1],
-                 top='top', bottom='bottom', alpha='alpha')
 
         fig.xaxis.minor_tick_line_alpha = 0
         fig.yaxis.visible = False
@@ -489,10 +491,6 @@ class ProcessingStacks(DashboardComponent):
             <span style="font-size: 14px; font-weight: bold;">Host:</span>&nbsp;
             <span style="font-size: 10px; font-family: Monaco, monospace;">@name</span>
         </div>
-        <div>
-            <span style="font-size: 14px; font-weight: bold;">Stacks:</span>&nbsp;
-            <span style="font-size: 10px; font-family: Monaco, monospace;">@stacks</span>
-        </div>
         <div>
             <span style="font-size: 14px; font-weight: bold;">Processing:</span>&nbsp;
             <span style="font-size: 10px; font-family: Monaco, monospace;">@processing</span>
@@ -510,12 +508,7 @@ class ProcessingStacks(DashboardComponent):
             data = self.processing_update(msg)
             x_range = self.root.x_range
             max_right = max(data['right'])
-            min_left = min(data['left'][:-1])
             cores = max(data['ncores'])
-            if min_left < x_range.start:  # not out there enough, jump ahead
-                x_range.start = min_left - 2
-            elif x_range.start < 2 * min_left - cores:  # way out there, walk back
-                x_range.start = x_range.start * 0.95 + min_left * 0.05
             if x_range.end < max_right:
                 x_range.end = max_right + 2
             elif x_range.end > 2 * max_right + cores:  # way out there, walk back
@@ -526,9 +519,7 @@ class ProcessingStacks(DashboardComponent):
     @staticmethod
     def processing_update(msg):
         with log_errors():
-            names = sorted(msg['stacks'])
-            stacks = msg['stacks']
-            stacks = [stacks[name] for name in names]
+            names = sorted(msg['processing'])
             names = sorted(names)
             processing = msg['processing']
             processing = [processing[name] for name in names]
@@ -537,22 +528,11 @@ class ProcessingStacks(DashboardComponent):
             n = len(names)
             d = {'name': list(names),
                  'processing': processing,
-                 'stacks': list(stacks),
                  'right': list(processing),
-                 'left': [-s for s in stacks],
                  'top': list(range(n, 0, -1)),
                  'bottom': list(range(n - 1, -1, -1)),
                  'ncores': ncores}
 
-            d['name'].append('ready')
-            d['processing'].append(0)
-            d['stacks'].append(msg['ready'])
-            d['left'].append(-msg['ready'] / n if n else 0)
-            d['right'].append(0)
-            d['top'].append(n)
-            d['bottom'].append(0)
-            d['ncores'].append(sum(msg['ncores'].values()))
-
-            d['alpha'] = [0.7] * n + [0.2]
+            d['alpha'] = [0.7] * n
 
             return d
