@@ -3,7 +3,6 @@ from __future__ import print_function, division, absolute_import
 import atexit
 import json
 import logging
-import multiprocessing
 import os
 import socket
 import subprocess
@@ -65,8 +64,12 @@ def main(host, port, http_port, bokeh_port, show, _bokeh,
     ip = socket.gethostbyname(host)
     loop = IOLoop.current()
     logger.info('-' * 47)
-    scheduler = Scheduler(ip=ip, loop=loop,
-                          services={('http', http_port): HTTPScheduler})
+
+    services = {('http', http_port): HTTPScheduler}
+    if _bokeh:
+        from distributed.bokeh.scheduler import BokehScheduler
+        services[('bokeh', 8788)] = BokehScheduler
+    scheduler = Scheduler(ip=ip, loop=loop, services=services)
     scheduler.start(port)
 
     bokeh_proc = None
@@ -82,7 +85,6 @@ def main(host, port, http_port, bokeh_port, show, _bokeh,
         except Exception as e:
             logger.warn("Could not start Bokeh web UI", exc_info=True)
 
-
     logger.info('-' * 47)
     try:
         loop.start()
