@@ -8,15 +8,18 @@ import shutil
 import socket
 from sys import argv, exit
 import sys
-from time import time, sleep
+from time import sleep
 
 import click
 from distributed import Nanny, Worker, sync, rpc
-from distributed.utils import get_ip, All
+from distributed.nanny import isalive
+from distributed.utils import get_ip, All, ignoring
 from distributed.worker import _ncores
 from distributed.http import HTTPWorker
+from distributed.metrics import time
 from distributed.cli.utils import check_python_3
-from tornado.ioloop import IOLoop
+from toolz import valmap
+from tornado.ioloop import IOLoop, TimeoutError
 from tornado import gen
 
 logger = logging.getLogger('distributed.dask_worker')
@@ -46,6 +49,9 @@ def handle_signal(sig, frame):
               help="Serving http port, defaults to randomly assigned")
 @click.option('--nanny-port', type=int, default=0,
               help="Serving nanny port, defaults to randomly assigned")
+@click.option('--bokeh-port', type=int, default=8789, help="Bokeh port")
+@click.option('--bokeh/--no-bokeh', 'bokeh', default=True, show_default=True,
+              required=False, help="Launch Bokeh Web UI")
 @click.option('--host', type=str, default=None,
               help="Serving host. Defaults to an ip address that can hopefully"
                    " be visible from the scheduler network.")
@@ -54,19 +60,26 @@ def handle_signal(sig, frame):
 @click.option('--nprocs', type=int, default=1,
               help="Number of worker processes.  Defaults to one.")
 @click.option('--name', type=str, default='', help="Alias")
-@click.option('--memory-limit', default=False,
+@click.option('--memory-limit', default='auto',
               help="Number of bytes before spilling data to disk. "
               "This can be an integer (nbytes) float (fraction of total memory) or auto")
-@click.option('--no-nanny', is_flag=True)
+@click.option('--reconnect/--no-reconnect', default=True,
+              help="Try to automatically reconnect to scheduler if disconnected")
+@click.option('--nanny/--no-nanny', default=True,
+              help="Start workers in nanny process for management")
 @click.option('--pid-file', type=str, default='',
               help="File to write the process PID")
-@click.option('--temp-filename', default=None, help="Internal use only")
+@click.option('--temp-filename', default=None,
+              help="Internal use only")
+@click.option('--resources', type=str, default='',
+              help='Resources for task constraints like "GPU=2 MEM=10e9"')
 def main(scheduler, host, worker_port, http_port, nanny_port, nthreads, nprocs,
-        no_nanny, name, memory_limit, pid_file, temp_filename):
-    if no_nanny:
-        port = worker_port
-    else:
+        nanny, name, memory_limit, pid_file, temp_filename, reconnect,
+        resources, bokeh, bokeh_port):
+    if nanny:
         port = nanny_port
+    else:
+        port = worker_port
 
     try:
         scheduler_host, scheduler_port = scheduler.split(':')
@@ -97,28 +110,31 @@ def main(scheduler, host, worker_port, http_port, nanny_port, nthreads, nprocs,
 
     services = {('http', http_port): HTTPWorker}
 
+    if bokeh:
+        try:
+            from distributed.bokeh.worker import BokehWorker
+        except ImportError:
+            pass
+        else:
+            services[('bokeh', bokeh_port)] = BokehWorker
+
+    if resources:
+        resources = resources.replace(',', ' ').split()
+        resources = dict(pair.split('=') for pair in resources)
+        resources = valmap(float, resources)
+    else:
+        resources = None
+
     loop = IOLoop.current()
 
-    if memory_limit == 'auto':
-        import psutil
-        memory_limit = psutil.virtual_memory().total * 0.60
-
-    if memory_limit:
-        memory_limit = float(memory_limit)
-        if memory_limit < 1.0:
-            import psutil
-            memory_limit = psutil.virtual_memory().total * memory_limit
-        memory_limit /= nprocs
-        memory_limit = int(memory_limit)
-
-    if no_nanny:
+    if nanny:
+        kwargs = {'worker_port': worker_port}
+        t = Nanny
+    else:
         kwargs = {}
         if nanny_port:
             kwargs['service_ports'] = {'nanny': nanny_port}
         t = Worker
-    else:
-        kwargs = {'worker_port': worker_port}
-        t = Nanny
 
     if host is not None:
         ip = socket.gethostbyname(host)
@@ -127,14 +143,14 @@ def main(scheduler, host, worker_port, http_port, nanny_port, nthreads, nprocs,
         # reach the scheduler
         ip = get_ip(scheduler_ip, scheduler_port)
     nannies = [t(scheduler_ip, scheduler_port, ncores=nthreads, ip=ip,
-                 services=services, name=name, loop=loop,
-                 memory_limit=memory_limit, **kwargs)
+                 services=services, name=name, loop=loop, resources=resources,
+                 memory_limit=memory_limit, reconnect=reconnect, **kwargs)
                for i in range(nprocs)]
 
-    for nanny in nannies:
-        nanny.start(port)
+    for n in nannies:
+        n.start(port)
         if t is Nanny:
-            global_nannies.append(nanny)
+            global_nannies.append(n)
 
     if temp_filename:
         @gen.coroutine
@@ -148,9 +164,18 @@ def main(scheduler, host, worker_port, http_port, nanny_port, nthreads, nprocs,
                 json.dump(msg, f)
         loop.add_callback(f)
 
-    loop.start()
-    logger.info("End worker")
-    loop.close()
+    @gen.coroutine
+    def run():
+        while all(n.status != 'closed' for n in nannies):
+            yield gen.sleep(0.2)
+
+    try:
+        loop.run_sync(run)
+    except (KeyboardInterrupt, TimeoutError):
+        pass
+    finally:
+        logger.info("End worker")
+        loop.close()
 
     loop2 = IOLoop()
 
@@ -158,20 +183,21 @@ def main(scheduler, host, worker_port, http_port, nanny_port, nthreads, nprocs,
     def f():
         scheduler = rpc(ip=nannies[0].scheduler.ip,
                         port=nannies[0].scheduler.port)
-        if not no_nanny:
+        if nanny:
             yield gen.with_timeout(timedelta(seconds=2),
                     All([scheduler.unregister(address=n.worker_address, close=True)
                         for n in nannies if n.process and n.worker_port]), io_loop=loop2)
 
     loop2.run_sync(f)
 
-    if not no_nanny:
+    if nanny:
         for n in nannies:
-            n.process.terminate()
+            if isalive(n.process):
+                n.process.terminate()
 
-    if not no_nanny:
+    if nanny:
         start = time()
-        while (any(n.process.poll() is None for n in nannies)
+        while (any(isalive(n.process) for n in nannies)
                 and time() < start + 1):
             sleep(0.1)
 
