@@ -9,12 +9,13 @@ import requests
 import signal
 import socket
 import sys
-from time import sleep, time
+from time import sleep
 
 
 from distributed import Scheduler, Client
 from distributed.utils import get_ip, ignoring, tmpfile
 from distributed.utils_test import loop, popen
+from distributed.metrics import time
 
 
 def test_defaults(loop):
@@ -53,11 +54,6 @@ def test_bokeh(loop):
         with Client('127.0.0.1:%d' % Scheduler.default_port, loop=loop) as c:
             pass
 
-        while True:
-            line = proc.stderr.readline()
-            if b'Web UI' in line:
-                break
-
         names = ['localhost', '127.0.0.1', get_ip()]
         if 'linux' in sys.platform:
             names.append(socket.gethostname())
@@ -87,11 +83,6 @@ def test_bokeh_non_standard_ports(loop):
         with Client('127.0.0.1:3448', loop=loop) as c:
             pass
 
-        while True:
-            line = proc.stderr.readline()
-            if b'Web UI' in line:
-                break
-
         start = time()
         while True:
             try:
@@ -102,7 +93,7 @@ def test_bokeh_non_standard_ports(loop):
                 sleep(0.1)
                 assert time() < start + 20
     with pytest.raises(Exception):
-        requests.get('http://127.0.0.1:4832/status/')
+        requests.get('http://localhost:4832/status/')
 
 
 @pytest.mark.skipif(not sys.platform.startswith('linux'),
@@ -117,11 +108,6 @@ def test_bokeh_whitelist(loop):
         with Client('127.0.0.1:%d' % Scheduler.default_port, loop=loop) as c:
             pass
 
-        while True:
-            line = proc.stderr.readline()
-            if b'Web UI' in line:
-                break
-
         start = time()
         while True:
             try:
@@ -137,8 +123,8 @@ def test_bokeh_whitelist(loop):
 
 def test_multiple_workers(loop):
     with popen(['dask-scheduler', '--no-bokeh']) as s:
-        with popen(['dask-worker', 'localhost:8786']) as a:
-            with popen(['dask-worker', 'localhost:8786']) as b:
+        with popen(['dask-worker', 'localhost:8786', '--no-bokeh']) as a:
+            with popen(['dask-worker', 'localhost:8786', '--no-bokeh']) as b:
                 with Client('127.0.0.1:%d' % Scheduler.default_port, loop=loop) as c:
                     start = time()
                     while len(c.ncores()) < 2:
@@ -164,9 +150,10 @@ def test_pid_file(loop):
             assert proc.pid == pid
 
     with tmpfile() as s:
-        with popen(['dask-scheduler', '--pid-file', s]) as sched:
+        with popen(['dask-scheduler', '--pid-file', s, '--no-bokeh']) as sched:
             check_pidfile(sched, s)
 
         with tmpfile() as w:
-            with popen(['dask-worker', '127.0.0.1:8786', '--pid-file', w]) as worker:
+            with popen(['dask-worker', '127.0.0.1:8786', '--pid-file', w,
+                        '--no-bokeh']) as worker:
                 check_pidfile(worker, w)
