@@ -1,6 +1,7 @@
 from __future__ import print_function, division, absolute_import
 
 from collections import defaultdict, Iterator, Iterable
+from concurrent.futures import ThreadPoolExecutor
 from concurrent.futures._base import DoneAndNotDoneFutures, CancelledError
 from contextlib import contextmanager
 import copy
@@ -72,17 +73,21 @@ class Future(WrappedKey):
     --------
     Client:  Creates futures
     """
+    _cb_executor = None
+    _cb_executor_pid = None
+
     def __init__(self, key, client):
         self.key = key
+        tkey = tokey(key)
         self.client = client
-        self.client._inc_ref(tokey(key))
+        self.client._inc_ref(tkey)
         self._generation = self.client.generation
         self._cleared = False
 
-        tkey = tokey(key)
-
-        if tkey not in client.futures:
-            client.futures[tkey] = {'event': Event(), 'status': 'pending'}
+        if tkey in client.futures:
+            self._state = client.futures[tkey]
+        else:
+            self._state = client.futures[tkey] = FutureState(Event())
 
     @property
     def executor(self):
@@ -90,14 +95,11 @@ class Future(WrappedKey):
 
     @property
     def status(self):
-        try:
-            return self.client.futures[tokey(self.key)]['status']
-        except KeyError:
-            return 'cancelled'
+        return self._state.status
 
     @property
     def event(self):
-        return self.client.futures[tokey(self.key)]['event']
+        return self._state.event
 
     def done(self):
         """ Is the computation complete? """
@@ -108,28 +110,27 @@ class Future(WrappedKey):
         result = sync(self.client.loop, self._result, raiseit=False)
         if self.status == 'error':
             six.reraise(*result)
-        if self.status == 'cancelled':
+        elif self.status == 'cancelled':
             raise result
         else:
             return result
 
     @gen.coroutine
     def _result(self, raiseit=True):
-        try:
-            d = self.client.futures[tokey(self.key)]
-        except KeyError:
+        yield self._state.event.wait()
+        if self.status == 'error':
+            exc = clean_exception(self._state.exception,
+                                  self._state.traceback)
+            if raiseit:
+                six.reraise(*exc)
+            else:
+                raise Return(exc)
+        elif self.status == 'cancelled':
             exception = CancelledError(self.key)
             if raiseit:
                 raise exception
             else:
                 raise gen.Return(exception)
-
-        yield d['event'].wait()
-        if self.status == 'error':
-            if raiseit:
-                six.reraise(*clean_exception(**d))
-            else:
-                raise Return(clean_exception(**d))
         else:
             result = yield self.client._gather([self])
             raise gen.Return(result[0])
@@ -138,8 +139,7 @@ class Future(WrappedKey):
     def _exception(self):
         yield self.event.wait()
         if self.status == 'error':
-            exception = self.client.futures[self.key]['exception']
-            raise Return(exception)
+            raise Return(self._state.exception)
         else:
             raise Return(None)
 
@@ -158,8 +158,22 @@ class Future(WrappedKey):
         The callback ``fn`` should take the future as its only argument.  This
         will be called regardless of if the future completes successfully,
         errs, or is cancelled
+
+        The callback is executed in a separate thread.
         """
-        self.client.loop.add_callback(done_callback, self, fn)
+        cls = Future
+        if cls._cb_executor is None or cls._cb_executor_pid != os.getpid():
+            cls._cb_executor = ThreadPoolExecutor(1)
+            cls._cb_executor_pid = os.getpid()
+
+        def execute_callback(fut):
+            try:
+                fn(fut)
+            except BaseException:
+                logger.exception("Error in callback %s of %s:", fn, fut)
+
+        self.client.loop.add_callback(done_callback, self,
+                                      partial(cls._cb_executor.submit, execute_callback))
 
     def cancel(self):
         """ Returns True if the future has been cancelled """
@@ -167,13 +181,13 @@ class Future(WrappedKey):
 
     def cancelled(self):
         """ Returns True if the future has been cancelled """
-        return tokey(self.key) not in self.client.futures
+        return self._state.status == 'cancelled'
 
     @gen.coroutine
     def _traceback(self):
         yield self.event.wait()
         if self.status == 'error':
-            raise Return(self.client.futures[tokey(self.key)]['traceback'])
+            raise Return(self._state.traceback)
         else:
             raise Return(None)
 
@@ -199,10 +213,7 @@ class Future(WrappedKey):
 
     @property
     def type(self):
-        try:
-            return self.client.futures[tokey(self.key)]['type']
-        except KeyError:
-            return None
+        return self._state.type
 
     def release(self):
         if not self._cleared and self.client.generation == self._generation:
@@ -235,9 +246,47 @@ class Future(WrappedKey):
     __repr__ = __str__
 
 
+class FutureState(object):
+    """A Future's internal state.
+
+    This is shared between all Futures with the same key and client.
+    """
+    __slots__ = ('event', 'status', 'type', 'exception', 'traceback')
+
+    def __init__(self, event):
+        self.event = event
+        self.status = 'pending'
+        self.type = None
+
+    def cancel(self):
+        self.status = 'cancelled'
+        self.event.set()
+
+    def finish(self, type=None):
+        self.status = 'finished'
+        self.event.set()
+        if type is not None:
+            self.type = type
+
+    def lose(self):
+        self.status = 'lost'
+        self.event.clear()
+
+    def set_error(self, exception, traceback):
+        self.status = 'error'
+        self.exception = exception
+        self.traceback = traceback
+        self.event.set()
+
+    def __str__(self):
+        return '<%s: %s>' % (self.__class__.__name__, self.status)
+
+    __repr__ = __str__
+
+
 @gen.coroutine
 def done_callback(future, callback):
-    """ Wait on future, then call callback """
+    """ Coroutine that waits on future, then calls callback """
     while future.status == 'pending':
         yield future.event.wait()
     callback(future)
@@ -248,6 +297,11 @@ def normalize_future(f):
     return [f.key, type(f)]
 
 
+class AllExit(Exception):
+    """Custom exception class to exit All(...) early.
+    """
+
+
 class Client(object):
     """ Drive computations on a distributed cluster
 
@@ -298,6 +352,8 @@ class Client(object):
         self.generation = 0
         self.status = None
         self._pending_msg_buffer = []
+        self.extensions = {}
+
         if hasattr(address, 'scheduler_address'):
             self.cluster = address
             address = address.scheduler_address
@@ -308,17 +364,31 @@ class Client(object):
             self._previous_shuffle = _globals.get('shuffle')
             dask.set_options(shuffle='tasks')
 
+        self._handlers = {
+                'key-in-memory': self._handle_key_in_memory,
+                'lost-data': self._handle_lost_data,
+                'cancelled-key': self._handle_cancelled_key,
+                'task-erred': self._handle_task_erred,
+                'restart': self._handle_restart,
+                'error': self._handle_error
+        }
+
         if start:
             self.start(timeout=timeout)
 
+        from distributed.channels import ChannelClient
+        ChannelClient(self)  # registers itself on construction
+
     def __str__(self):
         if hasattr(self, '_loop_thread'):
             n = sync(self.loop, self.scheduler.ncores)
-            return '<Client: scheduler="%s:%d" processes=%d cores=%d>' % (
+            return '<%s: scheduler="%s:%d" processes=%d cores=%d>' % (
+                    self.__class__.__name__,
                     self.scheduler.ip, self.scheduler.port, len(n),
                     sum(n.values()))
         else:
-            return '<Client: scheduler="%s:%d">' % (
+            return '<%s: scheduler="%s:%d">' % (
+                    self.__class__.__name__,
                     self.scheduler.ip, self.scheduler.port)
 
     __repr__ = __str__
@@ -376,10 +446,9 @@ class Client(object):
             self.status = 'connecting'
             self.scheduler_stream = None
 
-            events = [d['event'] for d in self.futures.values()]
+            for st in self.futures.values():
+                st.cancel()
             self.futures.clear()
-            for e in events:
-                e.set()
 
             while self.status == 'connecting':
                 try:
@@ -443,9 +512,9 @@ class Client(object):
     def _release_key(self, key):
         """ Release key from distributed memory """
         logger.debug("Release key %s", key)
-        if key in self.futures:
-            self.futures[key]['event'].clear()
-            del self.futures[key]
+        st = self.futures.pop(key, None)
+        if st is not None:
+            st.cancel()
         if self.status != 'closed':
             self._send_to_scheduler({'op': 'client-releases-keys',
                                      'keys': [key],
@@ -459,7 +528,7 @@ class Client(object):
                 try:
                     msgs = yield read(self.scheduler_stream.stream)
                 except StreamClosedError:
-                    logger.debug("Stream closed to scheduler", exc_info=True)
+                    logger.warn("Client report stream closed to scheduler")
                     if self.status == 'running':
                         logger.info("Reconnecting...")
                         self.status = 'connecting'
@@ -474,51 +543,61 @@ class Client(object):
                 for msg in msgs:
                     logger.debug("Client receives message %s", msg)
 
-                    if msg.get('status') == 'scheduler-error':
+                    if 'status' in msg and 'error' in msg['status']:
                         six.reraise(*clean_exception(**msg))
-                    elif msg['op'] == 'close':
+
+                    op = msg.pop('op')
+
+                    if op == 'close' or op == 'stream-closed':
                         breakout = True
                         break
-                    elif msg['op'] == 'key-in-memory':
-                        if msg['key'] in self.futures:
-                            self.futures[msg['key']]['status'] = 'finished'
-                            self.futures[msg['key']]['event'].set()
-                            if (msg.get('type') and
-                                not self.futures[msg['key']].get('type')):
-                                self.futures[msg['key']]['type'] = loads(msg['type'])
-                    elif msg['op'] == 'lost-data':
-                        if msg['key'] in self.futures:
-                            self.futures[msg['key']]['status'] = 'lost'
-                            self.futures[msg['key']]['event'].clear()
-                    elif msg['op'] == 'cancelled-key':
-                        if msg['key'] in self.futures:
-                            self.futures[msg['key']]['event'].set()
-                            del self.futures[msg['key']]
-                    elif msg['op'] == 'task-erred':
-                        if msg['key'] in self.futures:
-                            self.futures[msg['key']]['status'] = 'error'
-                            try:
-                                self.futures[msg['key']]['exception'] = loads(msg['exception'])
-                            except TypeError:
-                                self.futures[msg['key']]['exception'] = \
-                                    Exception('Undeserializable exception', msg['exception'])
-                            self.futures[msg['key']]['traceback'] = (loads(msg['traceback'])
-                                                                     if msg['traceback'] else None)
-                            self.futures[msg['key']]['event'].set()
-                    elif msg['op'] == 'restart':
-                        logger.info("Receive restart signal from scheduler")
-                        events = [d['event'] for d in self.futures.values()]
-                        self.futures.clear()
-                        for e in events:
-                            e.set()
-                        with ignoring(AttributeError):
-                            self._restart_event.set()
-                    elif 'error' in msg['op']:
-                        logger.warn("Scheduler exception:")
-                        logger.exception(msg['exception'])
+
+                    handler = self._handlers[op]
+                    handler(**msg)
                 if breakout:
                     break
 
+    def _handle_key_in_memory(self, key=None, type=None, workers=None):
+        state = self.futures.get(key)
+        if state is not None:
+            type = loads(type) if type and not state.type else None
+            state.finish(type)
+
+    def _handle_lost_data(self, key=None):
+        state = self.futures.get(key)
+        if state is not None:
+            state.lose()
+
+    def _handle_cancelled_key(self, key=None):
+        state = self.futures.get(key)
+        if state is not None:
+            state.cancel()
+
+    def _handle_task_erred(self, key=None, exception=None, traceback=None):
+        state = self.futures.get(key)
+        if state is not None:
+            try:
+                exception = loads(exception)
+            except TypeError:
+                exception = Exception("Undeserializable exception", exception)
+            if traceback:
+                traceback = loads(traceback)
+            else:
+                traceback = None
+            state.set_error(exception, traceback)
+
+    def _handle_restart(self):
+        logger.info("Receive restart signal from scheduler")
+        for state in self.futures.values():
+            state.cancel()
+        self.futures.clear()
+        with ignoring(AttributeError):
+            self._restart_event.set()
+
+    def _handle_error(self, exception=None):
+        logger.warn("Scheduler exception:")
+        logger.exception(exception)
+
     @gen.coroutine
     def _shutdown(self, fast=False):
         """ Send shutdown signal and wait until scheduler completes """
@@ -606,6 +685,7 @@ class Client(object):
         key = kwargs.pop('key', None)
         pure = kwargs.pop('pure', True)
         workers = kwargs.pop('workers', None)
+        resources = kwargs.pop('resources', None)
         allow_other_workers = kwargs.pop('allow_other_workers', False)
 
         if allow_other_workers not in (True, False, None):
@@ -640,7 +720,8 @@ class Client(object):
             dsk = {skey: (func,) + tuple(args)}
 
         futures = self._graph_to_futures(dsk, [skey], restrictions,
-                loose_restrictions, priority={skey: 0})
+                loose_restrictions, priority={skey: 0},
+                resources={skey: resources} if resources else None)
 
         logger.debug("Submit %s(...), %s", funcname(func), key)
 
@@ -716,6 +797,7 @@ class Client(object):
         key = key or funcname(func)
         pure = kwargs.pop('pure', True)
         workers = kwargs.pop('workers', None)
+        resources = kwargs.pop('resources', None)
         allow_other_workers = kwargs.pop('allow_other_workers', False)
 
         if allow_other_workers and workers is None:
@@ -763,8 +845,13 @@ class Client(object):
 
         priority = dict(zip(keys, range(len(keys))))
 
+        if resources:
+            resources = {key: resources for key in keys}
+        else:
+            resources = None
+
         futures = self._graph_to_futures(dsk, keys, restrictions,
-                loose_restrictions, priority=priority)
+                loose_restrictions, priority=priority, resources=resources)
         logger.debug("map(%s, ...)", funcname(func))
 
         return [futures[tokey(key)] for key in keys]
@@ -778,29 +865,32 @@ class Client(object):
         @gen.coroutine
         def wait(k):
             """ Want to stop the All(...) early if we find an error """
-            yield self.futures[k]['event'].wait()
-            if self.futures[k]['status'] != 'finished':
-                raise Exception()
+            st = self.futures[k]
+            yield st.event.wait()
+            if st.status != 'finished':
+                raise AllExit()
 
         while True:
             logger.debug("Waiting on futures to clear before gather")
 
-            with ignoring(Exception):
+            with ignoring(AllExit):
                 yield All([wait(key) for key in keys if key in self.futures])
 
+            failed = ('error', 'cancelled')
+
             exceptions = set()
             bad_keys = set()
             for key in keys:
                 if (key not in self.futures or
-                    self.futures[key]['status'] == 'error'):
+                    self.futures[key].status in failed):
                     exceptions.add(key)
                     if errors == 'raise':
                         try:
-                            d = self.futures[key]
-                            six.reraise(type(d['exception']),
-                                        d['exception'],
-                                        d['traceback'])
-                        except KeyError:
+                            st = self.futures[key]
+                            six.reraise(type(st.exception),
+                                        st.exception,
+                                        st.traceback)
+                        except (AttributeError, KeyError):
                             six.reraise(CancelledError,
                                         CancelledError(key),
                                         None)
@@ -814,11 +904,11 @@ class Client(object):
             response = yield self.scheduler.gather(keys=keys)
 
             if response['status'] == 'error':
-                logger.debug("Couldn't gather keys %s", response['keys'])
+                logger.warn("Couldn't gather keys %s", response['keys'])
                 self._send_to_scheduler({'op': 'missing-data',
                                          'keys': response['keys']})
                 for key in response['keys']:
-                    self.futures[key]['event'].clear()
+                    self.futures[key].event.clear()
             else:
                 break
 
@@ -916,15 +1006,14 @@ class Client(object):
                     "Input to scatter must be a list, iterator, or queue")
 
         for key in keys:
-            self.futures[key]['status'] = 'finished'
-            self.futures[key]['event'].set()
+            self.futures[key].finish(type=None)
 
         if isinstance(types, list):
             for key, typ in zip(keys, types):
-                self.futures[key]['type'] = typ
+                self.futures[key].type = typ
         elif isinstance(types, dict):
             for key in keys:
-                self.futures[key]['type'] = types[key]
+                self.futures[key].type = types[key]
 
         raise gen.Return(out)
 
@@ -1027,8 +1116,9 @@ class Client(object):
         keys = {tokey(f.key) for f in futures_of(futures)}
         yield self.scheduler.cancel(keys=list(keys), client=self.id)
         for k in keys:
-            with ignoring(KeyError):
-                del self.futures[k]
+            st = self.futures.pop(k, None)
+            if st is not None:
+                st.cancel()
 
     def cancel(self, futures):
         """
@@ -1046,13 +1136,14 @@ class Client(object):
 
     @gen.coroutine
     def _publish_dataset(self, **kwargs):
-        coroutines = []
-        for name, data in kwargs.items():
-            keys = [tokey(f.key) for f in futures_of(data)]
-            coroutines.append(self.scheduler.publish_dataset(keys=keys,
-                name=tokey(name), data=dumps(data), client=self.id))
+        with log_errors():
+            coroutines = []
+            for name, data in kwargs.items():
+                keys = [tokey(f.key) for f in futures_of(data)]
+                coroutines.append(self.scheduler.publish_put(keys=keys,
+                    name=tokey(name), data=dumps(data), client=self.id))
 
-        outs = yield coroutines
+            outs = yield coroutines
 
     def publish_dataset(self, **kwargs):
         """
@@ -1113,7 +1204,7 @@ class Client(object):
         --------
         Client.publish_dataset
         """
-        return sync(self.loop, self.scheduler.unpublish_dataset, name=name)
+        return sync(self.loop, self.scheduler.publist_delete, name=name)
 
     def list_datasets(self):
         """
@@ -1124,11 +1215,11 @@ class Client(object):
         Client.publish_dataset
         Client.get_dataset
         """
-        return sync(self.loop, self.scheduler.list_datasets)
+        return sync(self.loop, self.scheduler.publish_list)
 
     @gen.coroutine
     def _get_dataset(self, name):
-        out = yield self.scheduler.get_dataset(name=name, client=self.id)
+        out = yield self.scheduler.publish_get(name=name, client=self.id)
 
         with temp_default_client(self):
             data = loads(out['data'])
@@ -1159,7 +1250,7 @@ class Client(object):
             if resp['status'] == 'OK':
                 results[key] = resp['result']
             elif resp['status'] == 'error':
-                raise loads(resp['exception'])
+                six.reraise(*clean_exception(**resp))
         raise Return(results)
 
     def run(self, function, *args, **kwargs):
@@ -1197,8 +1288,53 @@ class Client(object):
         """
         return sync(self.loop, self._run, function, *args, **kwargs)
 
+    @gen.coroutine
+    def _run_coroutine(self, function, *args, **kwargs):
+        workers = kwargs.pop('workers', None)
+        wait = kwargs.pop('wait', True)
+        responses = yield self.scheduler.broadcast(msg=dict(op='run_coroutine',
+                                                function=dumps(function),
+                                                args=dumps(args),
+                                                kwargs=dumps(kwargs),
+                                                wait=wait),
+                                                workers=workers)
+        if not wait:
+            raise Return(None)
+        else:
+            results = {}
+            for key, resp in responses.items():
+                if resp['status'] == 'OK':
+                    results[key] = resp['result']
+                elif resp['status'] == 'error':
+                    six.reraise(*clean_exception(**resp))
+            raise Return(results)
+
+    def run_coroutine(self, function, *args, **kwargs):
+        """
+        Spawn a coroutine on all workers.
+
+        This spaws a coroutine on all currently known workers and then waits
+        for the coroutine on each worker.  The coroutines' results are returned
+        as a dictionary keyed by worker address.
+
+        Parameters
+        ----------
+        function: a coroutine function
+            (typically a function wrapped in gen.coroutine or
+             a Python 3.5+ async function)
+        *args: arguments for remote function
+        **kwargs: keyword arguments for remote function
+        wait: boolean (default True)
+            Whether to wait for coroutines to end.
+        workers: list
+            Workers on which to run the function. Defaults to all known workers.
+
+        """
+        return sync(self.loop, self._run_coroutine, function, *args, **kwargs)
+
     def _graph_to_futures(self, dsk, keys, restrictions=None,
-            loose_restrictions=None, allow_other_workers=True, priority=None):
+            loose_restrictions=None, allow_other_workers=True, priority=None,
+            resources=None):
 
         keyset = set(keys)
         flatkeys = list(map(tokey, keys))
@@ -1238,13 +1374,16 @@ class Client(object):
                                  'restrictions': restrictions or {},
                                  'loose_restrictions': loose_restrictions,
                                  'client': self.id,
-                                 'priority': priority})
+                                 'priority': priority,
+                                 'resources': resources})
 
         return futures
 
     @gen.coroutine
-    def _get(self, dsk, keys, restrictions=None, raise_on_error=True):
-        futures = self._graph_to_futures(dsk, set(flatten([keys])), restrictions)
+    def _get(self, dsk, keys, restrictions=None, raise_on_error=True,
+            resources=None):
+        futures = self._graph_to_futures(dsk, set(flatten([keys])),
+                restrictions, resources=resources)
 
         packed = pack_data(keys, futures)
         try:
@@ -1263,7 +1402,7 @@ class Client(object):
         raise gen.Return(result)
 
     def get(self, dsk, keys, restrictions=None, loose_restrictions=None,
-            **kwargs):
+            resources=None, **kwargs):
         """ Compute dask graph
 
         Parameters
@@ -1286,7 +1425,7 @@ class Client(object):
         Client.compute: Compute asynchronous collections
         """
         futures = self._graph_to_futures(dsk, set(flatten([keys])),
-                restrictions, loose_restrictions)
+                restrictions, loose_restrictions, resources=resources)
 
         try:
             results = self.gather(futures)
@@ -1352,7 +1491,7 @@ class Client(object):
             return redict_collection(collection, dsk)
 
     def compute(self, collections, sync=False, optimize_graph=True,
-            workers=None, allow_other_workers=False, **kwargs):
+            workers=None, allow_other_workers=False, resources=None, **kwargs):
         """ Compute dask collections on cluster
 
         Parameters
@@ -1417,8 +1556,12 @@ class Client(object):
         restrictions, loose_restrictions = get_restrictions(collections,
                 workers, allow_other_workers)
 
+        if resources:
+            resources = expand_resources(resources)
+
         futures_dict = self._graph_to_futures(merge(dsk2, dsk), names,
-                                              restrictions, loose_restrictions)
+                                              restrictions, loose_restrictions,
+                                              resources=resources)
 
         i = 0
         futures = []
@@ -1440,7 +1583,7 @@ class Client(object):
             return result
 
     def persist(self, collections, optimize_graph=True, workers=None,
-                allow_other_workers=None, **kwargs):
+                allow_other_workers=None, resources=None, **kwargs):
         """ Persist dask collections on cluster
 
         Starts computation of the collection on the cluster in the background.
@@ -1494,8 +1637,11 @@ class Client(object):
         restrictions, loose_restrictions = get_restrictions(collections,
                 workers, allow_other_workers)
 
+        if resources:
+            resources = expand_resources(resources)
+
         futures = self._graph_to_futures(dsk, names, restrictions,
-                loose_restrictions)
+                loose_restrictions, resources=resources)
 
         result = [redict_collection(c, {k: futures[k]
                                         for k in flatten(c._keys())})
@@ -1899,9 +2045,15 @@ class Client(object):
         """
         return sync(self.loop, self.scheduler.identity)
 
-    def get_versions(self):
+    def get_versions(self, check=False):
         """ Return version info for the scheduler, all workers and myself
 
+        Parameters
+        ----------
+        check : boolean, default False
+            raise ValueError if all required & optional packages
+            do not match
+
         Examples
         --------
         >>> c.get_versions()  # doctest: +SKIP
@@ -1922,7 +2074,36 @@ class Client(object):
                 return None
 
         workers = sync(self.loop, self._run, f)
-        return {'scheduler': scheduler, 'workers': workers, 'client': client}
+        result = {'scheduler': scheduler, 'workers': workers, 'client': client}
+
+        if check:
+            # we care about the required & optional packages matching
+            extract = lambda x: merge(result[x]['packages'].values())
+            client_versions = extract('client')
+            scheduler_versions = extract('scheduler')
+
+            for pkg, cv in client_versions.items():
+                sv = scheduler_versions[pkg]
+                if sv != cv:
+                    raise ValueError("package [{package}] is version [{client}] "
+                                     "on client and [{scheduler}] on scheduler!".format(
+                                         package=pkg,
+                                         client=cv,
+                                         scheduler=sv))
+
+            for w, d in workers.items():
+                worker_versions = merge(d['packages'].values())
+                for pkg, cv in client_versions.items():
+                    wv = worker_versions[pkg]
+                    if wv != cv:
+                        raise ValueError("package [{package}] is version [{client}] "
+                                         "on client and [{worker}] on worker [{w}]!".format(
+                                             package=pkg,
+                                             client=cv,
+                                             worker=wv,
+                                             w=w))
+
+        return result
 
     def futures_of(self, futures):
         return futures_of(futures, client=self)
@@ -2110,12 +2291,12 @@ def _wait(fs, timeout=None, return_when='ALL_COMPLETED'):
     if timeout is not None:
         raise NotImplementedError("Timeouts not yet supported")
     if return_when == 'ALL_COMPLETED':
-        try:
-            yield All({f.event.wait() for f in fs})
-        except KeyError:
-            raise CancelledError([f.key for f in fs
-                                        if f.key not in f.client.futures])
+        yield All({f.event.wait() for f in fs})
         done, not_done = set(fs), set()
+        cancelled = [f.key for f in done
+                     if f.status == 'cancelled']
+        if cancelled:
+            raise CancelledError(cancelled)
     else:
         raise NotImplementedError("Only return_when='ALL_COMPLETED' supported")
 
@@ -2179,6 +2360,8 @@ def as_completed(fs):
     This function does not return futures in the order in which they are input.
     """
     fs = list(fs)
+    if not fs:
+        return
     if len(set(f.client for f in fs)) == 1:
         loop = first(fs).client.loop
     else:
@@ -2263,6 +2446,22 @@ def temp_default_client(c):
     finally:
         _global_client[0] = old_exec
 
+
+def expand_resources(resources):
+    assert isinstance(resources, dict)
+    out = {}
+    for k, v in resources.items():
+        if not isinstance(k, tuple):
+            k = (k,)
+        for kk in k:
+            if hasattr(kk, '_keys'):
+                for kkk in kk._keys():
+                    out[tokey(kkk)] = v
+            else:
+                out[tokey(kk)] = v
+    return out
+
+
 def get_restrictions(collections, workers, allow_other_workers):
     """ Get restrictions from inputs to compute/persist """
     if isinstance(workers, (str, tuple, list)):
