@@ -1,6 +1,8 @@
 from __future__ import print_function, division, absolute_import
 
+from functools import partial
 import logging
+import math
 from threading import Thread
 from time import sleep
 
@@ -8,7 +10,6 @@ from tornado.ioloop import IOLoop
 from tornado.iostream import StreamClosedError
 from tornado import gen
 
-from ..http.scheduler import HTTPScheduler
 from ..utils import sync, ignoring, All
 from ..client import Client
 from ..nanny import Nanny
@@ -35,6 +36,11 @@ class LocalCluster(object):
         If False keep the workers in the main calling process
     scheduler_port: int
         Port of the scheduler.  8786 by default, use 0 to choose a random port
+    silence_logs: logging level
+        Level of logs to print out to stdout.  ``logging.CRITICAL`` by default.
+        Use a falsey value like False or None for no change.
+    kwargs: dict
+        Extra worker arguments, will be passed to the Worker constructor.
 
     Examples
     --------
@@ -56,9 +62,10 @@ class LocalCluster(object):
     def __init__(self, n_workers=None, threads_per_worker=None, nanny=True,
             loop=None, start=True, scheduler_port=8786,
             silence_logs=logging.CRITICAL, diagnostics_port=8787,
-            services={'http': HTTPScheduler}, **kwargs):
+            services={}, worker_services={}, **kwargs):
         self.status = None
         self.nanny = nanny
+        self.silence_logs = silence_logs
         if silence_logs:
             for l in ['distributed.scheduler',
                       'distributed.worker',
@@ -75,10 +82,11 @@ class LocalCluster(object):
         if n_workers is None and threads_per_worker is not None:
             n_workers = max(1, _ncores // threads_per_worker)
         if n_workers and threads_per_worker is None:
-            threads_per_worker = max(1, _ncores // n_workers)
+            # Overcommit threads per worker, rather than undercommit
+            threads_per_worker = max(1, int(math.ceil(_ncores / n_workers)))
 
         self.loop = loop or IOLoop()
-        if not self.loop._running:
+        if start and not self.loop._running:
             self._thread = Thread(target=self.loop.start)
             self._thread.daemon = True
             self._thread.start()
@@ -93,9 +101,10 @@ class LocalCluster(object):
         if start:
             _start_worker = self.start_worker
         else:
-            _start_worker = lambda *args, **kwargs: self.loop.add_callback(self._start_worker, *args, **kwargs)
+            _start_worker = partial(self.loop.add_callback, self._start_worker)
         for i in range(n_workers):
-            _start_worker(ncores=threads_per_worker, nanny=nanny)
+            _start_worker(ncores=threads_per_worker, nanny=nanny,
+                    services=worker_services, **kwargs)
         self.status = 'running'
 
         self.diagnostics = None
@@ -120,7 +129,8 @@ class LocalCluster(object):
             kwargs['quiet'] = True
         else:
             W = Worker
-        w = W(self.scheduler.ip, self.scheduler.port, loop=self.loop, **kwargs)
+        w = W(self.scheduler.ip, self.scheduler.port, loop=self.loop,
+              silence_logs=self.silence_logs, **kwargs)
         yield w._start(port)
 
         self.workers.append(w)
@@ -184,6 +194,7 @@ class LocalCluster(object):
         except ImportError:
             logger.info("To start diagnostics web server please install Bokeh")
             return
+        from ..http.scheduler import HTTPScheduler
 
         assert self.diagnostics is None
         if 'http' not in self.scheduler.services:
