@@ -2,15 +2,16 @@ from __future__ import print_function, division, absolute_import
 
 from functools import partial
 import sys
-from time import sleep, time
+from time import sleep
 import unittest
 
 from tornado.ioloop import IOLoop
 from tornado import gen
 import pytest
 
-from distributed.deploy.local import LocalCluster
 from distributed import Client, Worker, Nanny
+from distributed.deploy.local import LocalCluster
+from distributed.metrics import time
 from distributed.utils_test import inc, loop, raises, gen_test
 from distributed.utils import ignoring
 
@@ -189,6 +190,7 @@ def test_blocks_until_full(loop):
     with Client(loop=loop) as c:
         assert len(c.ncores()) > 0
 
+
 @gen_test()
 def test_scale_up_and_down():
     loop = IOLoop.current()
@@ -211,3 +213,13 @@ def test_scale_up_and_down():
 
     yield c._shutdown()
     yield cluster._close()
+
+
+def test_silent_startup(capsys, loop):
+    with LocalCluster(diagnostics_port=None, loop=loop, scheduler_port=0):
+        sleep(0.5)
+
+    out, err = capsys.readouterr()
+    assert not out
+    for line in err.split('\n'):
+        assert 'worker' not in line
