@@ -1,10 +1,15 @@
 from __future__ import print_function, division, absolute_import
 
 import os
-from time import time
+
+try:
+    from cytoolz import valmap
+except ImportError:
+    from toolz import valmap
 
 from toolz import countby, concat, dissoc
 
+from ..metrics import time
 from ..utils import key_split, log_errors
 
 
@@ -16,8 +21,6 @@ def tasks(s):
         return {'processing': processing,
                 'total': len(s.tasks),
                 'in-memory': len(s.who_has),
-                'ready': len(s.ready)
-                       + sum(map(len, s.stacks.values())),
                 'waiting': len(s.waiting),
                 'failed': len(s.exceptions_blame)}
 
@@ -39,7 +42,7 @@ def workers(s):
                    'network-send': 1234,
                    'network-recv': 1234,
                    'memory': 16701911040,
-                   'memory-percent': 85}}
+                   'memory_percent': 85}}
     """
     hosts = {host: ['%s:%s' % (host, port) for port in d['ports']]
                 for host, d in s.host_info.items()}
@@ -60,3 +63,10 @@ def workers(s):
             info['last-seen'] = (now - info['last-seen'])
 
     return result
+
+
+def processing(s):
+    return {'processing': valmap(len, s.processing),
+            'waiting': len(s.waiting),
+            'memory': len(s.who_has),
+            'ncores': s.ncores}
