@@ -74,7 +74,8 @@ def byte_sample(b, size, n):
         ends.append(min(start + size, starts[i + 1]))
     ends.append(starts[-1] + size)
 
-    return b''.join([b[start:end] for start, end in zip(starts, ends)])
+    parts = [b[start:end] for start, end in zip(starts, ends)]
+    return b''.join(map(ensure_bytes, parts))
 
 
 def maybe_compress(payload, compression=default_compression, min_size=1e4,
