@@ -1,9 +1,7 @@
 from __future__ import print_function, division, absolute_import
 
 from functools import partial
-from multiprocessing import Process
 import socket
-from time import time
 
 from tornado import gen, ioloop
 from tornado.iostream import StreamClosedError
@@ -11,7 +9,9 @@ import pytest
 
 from distributed.core import (read, write, pingpong, Server, rpc, connect,
         coerce_to_rpc, send_recv, coerce_to_address, ConnectionPool)
-from distributed.utils_test import slow, loop, gen_test
+from distributed.metrics import time
+from distributed.utils_test import slow, loop, gen_test, gen_cluster
+
 
 def test_server(loop):
     @gen.coroutine
@@ -24,7 +24,10 @@ def test_server(loop):
 
         stream = yield connect('127.0.0.1', 8887)
 
-        yield write(stream, {'op': 'ping'})
+        n = yield write(stream, {'op': 'ping'})
+        assert isinstance(n, int)
+        assert 4 <= n <= 1000
+
         response = yield read(stream)
         assert response == b'pong'
 
@@ -43,13 +46,15 @@ def test_rpc(loop):
         server = Server({'ping': pingpong})
         server.listen(8887)
 
-        remote = rpc(ip='127.0.0.1', port=8887)
+        with rpc(ip='127.0.0.1', port=8887) as remote:
+            response = yield remote.ping()
+            assert response == b'pong'
 
-        response = yield remote.ping()
-        assert response == b'pong'
+            response = yield remote.ping(close=True)
+            assert response == b'pong'
 
-        response = yield remote.ping(close=True)
-        assert response == b'pong'
+        assert not remote.streams
+        assert remote.status == 'closed'
 
         server.stop()
 
@@ -68,6 +73,9 @@ def test_rpc_inputs():
 
     assert all(r.ip == '127.0.0.1' and r.port == 8887 for r in L)
 
+    for r in L:
+        r.close_rpc()
+
 
 def test_rpc_with_many_connections(loop):
     remote = rpc(ip='127.0.0.1', port=8887)
@@ -123,11 +131,11 @@ def test_identity(loop):
         server = Server({})
         server.listen(8887)
 
-        remote = rpc(ip='127.0.0.1', port=8887)
-        a = yield remote.identity()
-        b = yield remote.identity()
-        assert a['type'] == 'Server'
-        assert a['id'] == b['id']
+        with rpc(ip='127.0.0.1', port=8887) as remote:
+            a = yield remote.identity()
+            b = yield remote.identity()
+            assert a['type'] == 'Server'
+            assert a['id'] == b['id']
 
         server.stop()
 
@@ -165,12 +173,12 @@ def test_errors(loop):
 
 
 def test_coerce_to_rpc():
-    r = coerce_to_rpc(('127.0.0.1', 8000))
-    assert (r.ip, r.port) == ('127.0.0.1', 8000)
-    r = coerce_to_rpc('127.0.0.1:8000')
-    assert (r.ip, r.port) == ('127.0.0.1', 8000)
-    r = coerce_to_rpc('foo:bar:8000')
-    assert (r.ip, r.port) == ('foo:bar', 8000)
+    with coerce_to_rpc(('127.0.0.1', 8000)) as r:
+        assert (r.ip, r.port) == ('127.0.0.1', 8000)
+    with coerce_to_rpc('127.0.0.1:8000') as r:
+        assert (r.ip, r.port) == ('127.0.0.1', 8000)
+    with coerce_to_rpc('foo:bar:8000') as r:
+        assert (r.ip, r.port) == ('foo:bar', 8000)
 
 
 def stream_div(stream=None, x=None, y=None):
@@ -181,16 +189,14 @@ def test_errors():
     server = Server({'div': stream_div})
     server.listen(0)
 
-    r = rpc(ip='127.0.0.1', port=server.port)
-    with pytest.raises(ZeroDivisionError):
-        yield r.div(x=1, y=0)
-
-    r.close_streams()
+    with rpc(ip='127.0.0.1', port=server.port) as r:
+        with pytest.raises(ZeroDivisionError):
+            yield r.div(x=1, y=0)
 
 
 @gen_test()
 def test_connect_raises():
-    with pytest.raises((gen.TimeoutError, StreamClosedError)):
+    with pytest.raises((gen.TimeoutError, StreamClosedError, IOError)):
         yield connect('127.0.0.1', 58259, timeout=0.01)
 
 
@@ -261,3 +267,12 @@ def test_connection_pool():
     while any(rpc.available.values()):
         yield gen.sleep(0.01)
         assert time() < start + 2
+
+
+@gen_cluster()
+def test_ticks(s, a, b):
+    pytest.importorskip('crick')
+    yield gen.sleep(0.1)
+    c = s.digests['tick-duration']
+    assert c.size()
+    assert 0.01 < c.components[0].quantile(0.5) < 0.5
