@@ -4,9 +4,9 @@ import cloudpickle
 from collections import defaultdict, deque
 from copy import deepcopy
 from datetime import timedelta
-from operator import add
+from operator import add, mul
 import sys
-from time import time, sleep
+from time import sleep
 
 import dask
 from dask import delayed
@@ -20,23 +20,23 @@ from tornado import gen
 import pytest
 
 from distributed import Nanny, Worker
-from distributed.batched import BatchedStream
-from distributed.core import connect, read, write, rpc
-from distributed.scheduler import (validate_state, decide_worker,
-        Scheduler)
-from distributed.client import _wait
+from distributed.core import connect, read, write, close, rpc
+from distributed.scheduler import validate_state, Scheduler, BANDWIDTH
+from distributed.client import _wait, _first_completed
+from distributed.metrics import time
 from distributed.protocol.pickle import dumps
 from distributed.worker import dumps_function, dumps_task
 from distributed.utils_test import (inc, ignoring, dec, gen_cluster, gen_test,
-        loop)
+        loop, readone, slowinc, slowadd)
 from distributed.utils import All
+from distributed.utils_test import slow
 from dask.compatibility import apply
 
 
 alice = 'alice:1234'
 bob = 'bob:1234'
 
-stack_duration = defaultdict(lambda: 0)
+occupancy = defaultdict(lambda: 0)
 
 
 @gen_cluster()
@@ -48,88 +48,8 @@ def test_administration(s, a, b):
     assert str(len(s.ncores)) in repr(s)
 
 
-@gen_cluster(ncores=[])
-def test_update_state(s):
-    s.add_worker(address=alice, ncores=1, coerce_address=False)
-    s.update_graph(tasks={'x': 1, 'y': dumps_task((inc, 'x'))},
-                   keys=['y'],
-                   dependencies={'y': 'x', 'x': set()},
-                   client='client')
-
-    s.ensure_occupied()
-    r = s.transition('x', 'memory', nbytes=10, type=dumps(int),
-            compute_start=10, compute_stop=11, worker=alice)
-    s.transitions(r)
-    s.ensure_occupied()
-
-    assert set(s.processing[alice]) == {'y'}
-    assert set(s.rprocessing['y']) == {alice}
-    assert not s.ready
-    assert s.who_wants == {'y': {'client'}}
-    assert s.wants_what == {'client': {'y'}}
-
-    s.update_graph(tasks={'a': 1, 'z': dumps_task((add, 'y', 'a'))},
-                   keys=['z'],
-                   dependencies={'z': {'y', 'a'}},
-                   client='client')
-
-
-    assert s.tasks == {'x': 1, 'y': dumps_task((inc, 'x')), 'a': 1,
-                       'z': dumps_task((add, 'y', 'a'))}
-    assert s.dependencies == {'x': set(), 'a': set(), 'y': {'x'}, 'z': {'a', 'y'}}
-    assert s.dependents == {'z': set(), 'y': {'z'}, 'a': {'z'}, 'x': {'y'}}
-
-    assert s.waiting == {'z': {'a', 'y'}}
-    assert s.waiting_data == {'x': {'y'}, 'y': {'z'}, 'a': {'z'}, 'z': set()}
-
-    assert s.who_wants == {'z': {'client'}, 'y': {'client'}}
-    assert s.wants_what == {'client': {'y', 'z'}}
-
-    assert 'a' in s.ready or 'a' in s.processing[alice]
-
-
-@gen_cluster(ncores=[])
-def test_update_state_with_processing(s):
-    s.add_worker(address=alice, ncores=1, coerce_address=False)
-    s.update_graph(tasks={'x': 1, 'y': dumps_task((inc, 'x')),
-                          'z': dumps_task((inc, 'y'))},
-                   keys=['z'],
-                   dependencies={'y': {'x'}, 'x': set(), 'z': {'y'}},
-                   client='client')
-
-    s.ensure_occupied()
-    r = s.transition('x', 'memory', nbytes=10, type=dumps(int),
-            compute_start=10, compute_stop=11, worker=alice)
-    s.transitions(r)
-    s.ensure_occupied()
-
-    assert s.waiting == {'z': {'y'}}
-    assert s.waiting_data == {'x': {'y'}, 'y': {'z'}, 'z': set()}
-    assert list(s.ready) == []
-
-    assert s.who_wants == {'z': {'client'}}
-    assert s.wants_what == {'client': {'z'}}
-
-    assert s.who_has == {'x': {alice}}
-
-    s.update_graph(tasks={'a': dumps_task((inc, 'x')), 'b': (add,'a','y'),
-                          'c': dumps_task((inc, 'z'))},
-                   keys=['b', 'c'],
-                   dependencies={'a': {'x'}, 'b': {'a', 'y'}, 'c': {'z'}},
-                   client='client')
-
-    assert s.waiting == {'z': {'y'}, 'b': {'a', 'y'}, 'c': {'z'}}
-    assert 'a' in s.stacks[alice] or 'a' in s.processing[alice]
-    assert not s.ready
-    assert s.waiting_data == {'x': {'y', 'a'}, 'y': {'z', 'b'}, 'z': {'c'},
-                              'a': {'b'}, 'b': set(), 'c': set()}
-
-    assert s.who_wants == {'b': {'client'}, 'c': {'client'}, 'z': {'client'}}
-    assert s.wants_what == {'client': {'b', 'c', 'z'}}
-
-
 @gen_cluster(client=True, ncores=[('127.0.0.1', 1)])
-def test_update_state_respects_data_in_memory(c, s, a):
+def test_respect_data_in_memory(c, s, a):
     x = delayed(inc)(1)
     y = delayed(inc)(x)
     f = c.persist(y)
@@ -145,132 +65,98 @@ def test_update_state_respects_data_in_memory(c, s, a):
         yield gen.sleep(0.0001)
 
 
-@gen_cluster(ncores=[])
-def test_update_state_supports_recomputing_released_results(s):
-    s.add_worker(address=alice, ncores=1, coerce_address=False)
-    s.update_graph(tasks={'x': 1, 'y': dumps_task((inc, 'x')),
-                          'z': dumps_task((inc, 'x'))},
-                   keys=['z'],
-                   dependencies={'y': {'x'}, 'x': set(), 'z': {'y'}},
-                   client='client')
+@gen_cluster(client=True)
+def test_recompute_released_results(c, s, a, b):
+    x = delayed(inc)(1)
+    y = delayed(inc)(x)
 
-    s.ensure_occupied()
-    r = s.transition('x', 'memory', nbytes=10, type=dumps(int),
-            compute_start=10, compute_stop=11, worker=alice)
-    s.transitions(r)
-    s.ensure_occupied()
-    r = s.transition('y', 'memory', nbytes=10, type=dumps(int),
-            compute_start=10, compute_stop=11, worker=alice)
-    s.transitions(r)
-    s.ensure_occupied()
-    r = s.transition('z', 'memory', nbytes=10, type=dumps(int),
-            compute_start=10, compute_stop=11, worker=alice)
-    s.transitions(r)
-    s.ensure_occupied()
+    yy = c.persist(y)
+    yield _wait(yy)
 
-    assert not s.waiting
-    assert not s.ready
-    assert s.waiting_data == {'z': set()}
+    while x.key in s.who_has or x.key in a.data or x.key in b.data:  # let x go away
+        yield gen.sleep(0.01)
 
-    assert s.who_has == {'z': {alice}}
-
-    s.update_graph(tasks={'x': 1, 'y': dumps_task((inc, 'x'))},
-                   keys=['y'],
-                   dependencies={'y': {'x'}, 'x': set()},
-                   client='client')
-
-    assert s.waiting == {'y': {'x'}}
-    assert s.waiting_data == {'x': {'y'}, 'y': set(), 'z': set()}
-    assert s.who_wants == {'z': {'client'}, 'y': {'client'}}
-    assert s.wants_what == {'client': {'y', 'z'}}
-    assert set(s.processing[alice]) == {'x'}
+    z = delayed(dec)(x)
+    zz = c.compute(z)
+    result = yield zz._result()
+    assert result == 1
 
 
-def test_decide_worker_with_many_independent_leaves():
-    dsk = merge({('y', i): (inc, ('x', i)) for i in range(100)},
-                {('x', i): i for i in range(100)})
-    dependencies, dependents = get_deps(dsk)
-    stacks = {alice: [], bob: []}
-    processing = {alice: dict(), bob: dict()}
-    who_has = merge({('x', i * 2): {alice} for i in range(50)},
-                    {('x', i * 2 + 1): {bob} for i in range(50)})
-    nbytes = {k: 0 for k in who_has}
-    ncores = {alice: 1, bob: 1}
+@gen_cluster(client=True)
+def test_decide_worker_with_many_independent_leaves(c, s, a, b):
+    xs = yield [c._scatter(list(range(0, 100, 2)), workers=a.address),
+                c._scatter(list(range(1, 100, 2)), workers=b.address)]
+    xs = list(concat(zip(*xs)))
+    ys = [delayed(inc)(x) for x in xs]
 
-    for key in dsk:
-        worker = decide_worker(dependencies, stacks, stack_duration, processing,
-                               who_has, {}, {}, set(), nbytes, ncores, key)
-        stacks[worker].append(key)
+    y2s = c.persist(ys)
+    yield _wait(y2s)
 
-    nhits = (len([k for k in stacks[alice] if alice in who_has[('x', k[1])]])
-             + len([k for k in stacks[bob] if bob in who_has[('x', k[1])]]))
+    nhits = (sum(y.key in a.data for y in y2s[::2]) +
+             sum(y.key in b.data for y in y2s[1::2]))
 
-    assert nhits > 90
+    assert nhits > 80
 
 
-def test_decide_worker_with_restrictions():
-    dependencies = {'x': set()}
-    alice, bob, charlie = 'alice:8000', 'bob:8000', 'charlie:8000'
-    stacks = {alice: [], bob: [], charlie: []}
-    processing = {alice: dict(), bob: dict(), charlie: dict()}
-    who_has = {}
-    restrictions = {'x': {'alice', 'charlie'}}
-    nbytes = {}
-    ncores = {alice: 1, bob: 1, charlie: 1}
-    result = decide_worker(dependencies, stacks, stack_duration, processing,
-                          who_has, {}, restrictions, set(), nbytes, ncores, 'x')
-    assert result in {alice, charlie}
+@gen_cluster(client=True, ncores=[('127.0.0.1', 1)] * 3)
+def test_decide_worker_with_restrictions(client, s, a, b, c):
+    x = client.submit(inc, 1, workers=[a.address, b.address])
+    yield _wait(x)
+    assert x.key in a.data or x.key in b.data
 
 
-    stacks = {alice: [1, 2, 3], bob: [], charlie: [4, 5, 6]}
-    result = decide_worker(dependencies, stacks, stack_duration, processing,
-                          who_has, {}, restrictions, set(), nbytes, ncores, 'x')
-    assert result in {alice, charlie}
-
-    dependencies = {'x': {'y'}}
-    who_has = {'y': {bob}}
-    nbytes = {'y': 0}
-    result = decide_worker(dependencies, stacks, stack_duration, processing,
-                          who_has, {}, restrictions, set(), nbytes, ncores, 'x')
-    assert result in {alice, charlie}
+@gen_cluster(client=True, ncores=[('127.0.0.1', 1)] * 3)
+def test_move_data_over_break_restrictions(client, s, a, b, c):
+    [x] = yield client._scatter([1], workers=b.address)
+    y = client.submit(inc, x, workers=[a.address, b.address])
+    yield _wait(y)
+    assert y.key in a.data or y.key in b.data
 
 
-def test_decide_worker_with_loose_restrictions():
-    dependencies = {'x': set()}
-    alice, bob, charlie = 'alice:8000', 'bob:8000', 'charlie:8000'
-    stacks = {alice: [1, 2, 3], bob: [], charlie: [1]}
-    stack_duration = {alice: 3, bob: 0, charlie: 1}
-    processing = {alice: dict(), bob: dict(), charlie: dict()}
-    who_has = {}
-    nbytes = {}
-    ncores = {alice: 1, bob: 1, charlie: 1}
-    restrictions = {'x': {'alice', 'charlie'}}
+@gen_cluster(client=True, ncores=[('127.0.0.1', 1)] * 3)
+def test_balance_with_restrictions(client, s, a, b, c):
+    [x], [y] = yield [client._scatter([[1, 2, 3]], workers=a.address),
+                      client._scatter([1], workers=c.address)]
+    z = client.submit(inc, 1, workers=[a.address, c.address])
+    yield _wait(z)
 
-    result = decide_worker(dependencies, stacks, stack_duration, processing,
-                          who_has, {}, restrictions, set(), nbytes, ncores, 'x')
-    assert result == charlie
-
-    result = decide_worker(dependencies, stacks, stack_duration, processing,
-                          who_has, {}, restrictions, {'x'}, nbytes, ncores, 'x')
-    assert result == charlie
-
-    restrictions = {'x': {'david', 'ethel'}}
-    result = decide_worker(dependencies, stacks, stack_duration, processing,
-                          who_has, {}, restrictions, set(), nbytes, ncores, 'x')
-    assert result is None
-
-    restrictions = {'x': {'david', 'ethel'}}
-    result = decide_worker(dependencies, stacks, stack_duration, processing,
-                          who_has, {}, restrictions, {'x'}, nbytes, ncores, 'x')
-
-    assert result == bob
+    assert s.who_has[z.key] == {c.address}
 
 
-def test_decide_worker_without_stacks():
-    assert not decide_worker({'x': []}, {}, {}, {}, {}, {}, {}, set(), {}, {},
-                             'x')
+@gen_cluster(client=True, ncores=[('127.0.0.1', 1)] * 3)
+def test_no_valid_workers(client, s, a, b, c):
+    x = client.submit(inc, 1, workers='127.0.0.5:9999')
+    while not s.tasks:
+        yield gen.sleep(0.01)
+
+    assert x.key in s.unrunnable
+
+    with pytest.raises(gen.TimeoutError):
+        yield gen.with_timeout(timedelta(milliseconds=50), x._result())
 
 
+@gen_cluster(client=True, ncores=[('127.0.0.1', 1)] * 3)
+def test_no_valid_workers_loose_restrictions(client, s, a, b, c):
+    x = client.submit(inc, 1, workers='127.0.0.5:9999',
+                      allow_other_workers=True)
+
+    result = yield x._result()
+    assert result == 2
+
+
+@gen_cluster(client=True, ncores=[])
+def test_no_workers(client, s):
+    x = client.submit(inc, 1)
+    while not s.tasks:
+        yield gen.sleep(0.01)
+
+    assert x.key in s.unrunnable
+
+    with pytest.raises(gen.TimeoutError):
+        yield gen.with_timeout(timedelta(milliseconds=50), x._result())
+
+
+@pytest.mark.skip
 def test_validate_state():
     dsk = {'x': 1, 'y': (inc, 'x')}
     dependencies = {'x': set(), 'y': {'x'}}
@@ -338,143 +224,32 @@ def test_validate_state():
     validate_state(**locals())
 
 
-def div(x, y):
-    return x / y
-
-@gen_cluster()
-def test_scheduler(s, a, b):
-    stream = yield connect(s.ip, s.port)
-    yield write(stream, {'op': 'register-client', 'client': 'ident'})
-    stream = BatchedStream(stream, 10)
-    msg = yield read(stream)
-    assert msg['op'] == 'stream-start'
-
-    # Test update graph
-    yield write(stream, {'op': 'update-graph',
-                         'tasks': valmap(dumps_task, {'x': (inc, 1),
-                                                      'y': (inc, 'x'),
-                                                      'z': (inc, 'y')}),
-                         'dependencies': {'x': [],
-                                          'y': ['x'],
-                                          'z': ['y']},
-                         'keys': ['x', 'z'],
-                         'client': 'ident'})
-    while True:
-        msg = yield read(stream)
-        if msg['op'] == 'key-in-memory' and msg['key'] == 'z':
-            break
-
-    assert a.data.get('x') == 2 or b.data.get('x') == 2
-
-    # Test erring tasks
-    yield write(stream, {'op': 'update-graph',
-                         'tasks': valmap(dumps_task, {'a': (div, 1, 0),
-                                                       'b': (inc, 'a')}),
-                         'dependencies': {'a': [],
-                                           'b': ['a']},
-                         'keys': ['a', 'b'],
-                         'client': 'ident'})
-
-    while True:
-        msg = yield read(stream)
-        if msg['op'] == 'task-erred' and msg['key'] == 'b':
-            break
-
-    # Test missing data
-    yield write(stream, {'op': 'missing-data', 'keys': ['z']})
-    s.ensure_occupied()
-
-    while True:
-        msg = yield read(stream)
-        if msg['op'] == 'key-in-memory' and msg['key'] == 'z':
-            break
-
-    # Test missing data without being informed
-    for w in [a, b]:
-        if 'z' in w.data:
-            del w.data['z']
-    yield write(stream, {'op': 'update-graph',
-                         'tasks': {'zz': dumps_task((inc, 'z'))},
-                         'dependencies': {'zz': ['z']},
-                         'keys': ['zz'],
-                         'client': 'ident'})
-    while True:
-        msg = yield read(stream)
-        if msg['op'] == 'key-in-memory' and msg['key'] == 'zz':
-            break
-
-    write(stream, {'op': 'close'})
-    stream.close()
-
-
-@gen_cluster()
-def test_multi_queues(s, a, b):
-    sched, report = Queue(), Queue()
-    s.handle_queues(sched, report)
-
-    msg = yield report.get()
-    assert msg['op'] == 'stream-start'
-
-    # Test update graph
-    sched.put_nowait({'op': 'update-graph',
-                      'tasks': valmap(dumps_task, {'x': (inc, 1),
-                                                   'y': (inc, 'x'),
-                                                   'z': (inc, 'y')}),
-                      'dependencies': {'x': [],
-                                       'y': ['x'],
-                                       'z': ['y']},
-                      'keys': ['z']})
-
-    while True:
-        msg = yield report.get()
-        if msg['op'] == 'key-in-memory' and msg['key'] == 'z':
-            break
-
-    slen, rlen = len(s.scheduler_queues), len(s.report_queues)
-    sched2, report2 = Queue(), Queue()
-    s.handle_queues(sched2, report2)
-    assert slen + 1 == len(s.scheduler_queues)
-    assert rlen + 1 == len(s.report_queues)
-
-    sched2.put_nowait({'op': 'update-graph',
-                       'tasks': {'a': dumps_task((inc, 10))},
-                       'dependencies': {'a': []},
-                       'keys': ['a']})
-
-    for q in [report, report2]:
-        while True:
-            msg = yield q.get()
-            if msg['op'] == 'key-in-memory' and msg['key'] == 'a':
-                break
-
-
 @gen_cluster()
 def test_server(s, a, b):
     stream = yield connect('127.0.0.1', s.port)
     yield write(stream, {'op': 'register-client', 'client': 'ident'})
-    stream = BatchedStream(stream, 0)
-    stream.send({'op': 'update-graph',
-                 'tasks': {'x': dumps_task((inc, 1)),
-                           'y': dumps_task((inc, 'x'))},
-                 'dependencies': {'x': [], 'y': ['x']},
-                 'keys': ['y'],
-                 'client': 'ident'})
+    yield write(stream, {'op': 'update-graph',
+                         'tasks': {'x': dumps_task((inc, 1)),
+                                   'y': dumps_task((inc, 'x'))},
+                         'dependencies': {'x': [], 'y': ['x']},
+                         'keys': ['y'],
+                         'client': 'ident'})
 
     while True:
-        msg = yield read(stream)
+        msg = yield readone(stream)
         if msg['op'] == 'key-in-memory' and msg['key'] == 'y':
             break
 
-    stream.send({'op': 'close-stream'})
-    msg = yield read(stream)
+    yield write(stream, {'op': 'close-stream'})
+    msg = yield readone(stream)
     assert msg == {'op': 'stream-closed'}
-    assert stream.closed()
-    stream.close()
+    with pytest.raises(StreamClosedError):
+        yield readone(stream)
+    close(stream)
 
 
 @gen_cluster()
 def test_remove_client(s, a, b):
-    s.add_client(client='ident')
     s.update_graph(tasks={'x': dumps_task((inc, 1)),
                           'y': dumps_task((inc, 'x'))},
                    dependencies={'x': [], 'y': ['x']},
@@ -499,17 +274,14 @@ def test_server_listens_to_other_ops(s, a, b):
 
 @gen_cluster()
 def test_remove_worker_from_scheduler(s, a, b):
-    dsk = {('x', i): (inc, i) for i in range(20)}
+    dsk = {('x-%d' % i): (inc, i) for i in range(20)}
     s.update_graph(tasks=valmap(dumps_task, dsk), keys=list(dsk),
                    dependencies={k: set() for k in dsk})
-    assert s.ready
-    assert not any(stack for stack in s.stacks.values())
 
     assert a.address in s.worker_streams
     s.remove_worker(address=a.address)
     assert a.address not in s.ncores
-    assert len(s.ready) + len(s.processing[b.address]) == \
-            len(dsk)  # b owns everything
+    assert len(s.processing[b.address]) == len(dsk)  # b owns everything
     s.validate_state()
 
 
@@ -520,7 +292,7 @@ def test_add_worker(s, a, b):
     w.data['y'] = 1
     yield w._start(0)
 
-    dsk = {('x-%d' % i).encode(): (inc, i) for i in range(10)}
+    dsk = {('x-%d' % i): (inc, i) for i in range(10)}
     s.update_graph(tasks=valmap(dumps_task, dsk), keys=list(dsk), client='client',
                    dependencies={k: set() for k in dsk})
 
@@ -537,7 +309,7 @@ def test_add_worker(s, a, b):
 @gen_cluster()
 def test_feed(s, a, b):
     def func(scheduler):
-        return dumps((scheduler.processing, scheduler.stacks))
+        return dumps(scheduler.processing)
 
     stream = yield connect(s.ip, s.port)
     yield write(stream, {'op': 'feed',
@@ -546,10 +318,10 @@ def test_feed(s, a, b):
 
     for i in range(5):
         response = yield read(stream)
-        expected = s.processing, s.stacks
+        expected = s.processing
         assert cloudpickle.loads(response) == expected
 
-    stream.close()
+    close(stream)
 
 
 @gen_cluster()
@@ -575,7 +347,7 @@ def test_feed_setup_teardown(s, a, b):
         response = yield read(stream)
         assert response == 'OK'
 
-    stream.close()
+    close(stream)
     start = time()
     while not hasattr(s, 'flag'):
         yield gen.sleep(0.01)
@@ -601,7 +373,7 @@ def test_feed_large_bytestring(s, a, b):
         response = yield read(stream)
         assert response == True
 
-    stream.close()
+    close(stream)
 
 
 @gen_test(timeout=None)
@@ -653,44 +425,18 @@ def test_delete_data(c, s, a, b):
         assert time() < start + 5
 
 
-@gen_cluster(client=True)
-def test_delete_callback(c, s, a, b):
-    d = yield c._scatter({'x': 1}, workers=a.address)
+@gen_cluster(client=True, ncores=[('127.0.0.1', 1)])
+def test_delete(c, s, a):
+    x = c.submit(inc, 1)
+    yield x._result()
+    assert x.key in a.data
 
-    assert s.who_has['x'] == {a.address}
-    assert s.has_what[a.address] == {'x'}
+    yield c._cancel(x)
 
-    s.client_releases_keys(client=c.id, keys=['x'])
-
-    assert 'x' not in s.who_has
-    assert 'x' not in s.has_what[a.address]
-    assert a.data['x'] == 1  # still in memory
-    assert s.deleted_keys == {a.address: {'x'}}
-    yield s.clear_data_from_workers()
-    assert not s.deleted_keys
-    assert not a.data
-
-    assert s._delete_periodic_callback.is_running()
-
-
-@gen_cluster()
-def test_self_aliases(s, a, b):
-    a.data['a'] = 1
-    s.update_data(who_has={'a': [a.address]},
-                  nbytes={'a': 10}, client='client')
-    s.update_graph(tasks=valmap(dumps_task, {'a': 'a', 'b': (inc, 'a')}),
-                   keys=['b'], client='client',
-                   dependencies={'b': ['a']})
-
-    sched, report = Queue(), Queue()
-    s.handle_queues(sched, report)
-    msg = yield report.get()
-    assert msg['op'] == 'stream-start'
-
-    while True:
-        msg = yield report.get()
-        if msg['op'] == 'key-in-memory' and msg['key'] == 'b':
-            break
+    start = time()
+    while x.key in a.data:
+        yield gen.sleep(0.01)
+        assert time() < start + 5
 
 
 @gen_cluster()
@@ -701,8 +447,6 @@ def test_filtered_communication(s, a, b):
     yield write(f, {'op': 'register-client', 'client': 'f'})
     yield read(c)
     yield read(f)
-    c = BatchedStream(c, 0)
-    f = BatchedStream(f, 0)
 
     assert set(s.streams) == {'c', 'f'}
 
@@ -720,10 +464,10 @@ def test_filtered_communication(s, a, b):
                     'client': 'f',
                     'keys': ['z']})
 
-    msg = yield read(c)
+    msg, = yield read(c)
     assert msg['op'] == 'key-in-memory'
     assert msg['key'] == 'y'
-    msg = yield read(f)
+    msg, = yield read(f)
     assert msg['op'] == 'key-in-memory'
     assert msg['key'] == 'z'
 
@@ -757,7 +501,6 @@ def test_dumps_task():
 
 @gen_cluster()
 def test_ready_remove_worker(s, a, b):
-    s.add_client(client='client')
     s.update_graph(tasks={'x-%d' % i: dumps_task((inc, i)) for i in range(20)},
                    keys=['x-%d' % i for i in range(20)],
                    client='client',
@@ -765,68 +508,33 @@ def test_ready_remove_worker(s, a, b):
 
     assert all(len(s.processing[w]) >= s.ncores[w]
                 for w in s.ncores)
-    assert not any(stack for stack in s.stacks.values())
-    assert len(s.ready) + sum(map(len, s.processing.values())) == 20
 
     s.remove_worker(address=a.address)
 
-    for collection in [s.ncores, s.stacks, s.processing]:
+    for collection in [s.ncores, s.processing]:
         assert set(collection) == {b.address}
     assert all(len(s.processing[w]) >= s.ncores[w]
                 for w in s.ncores)
     assert set(s.processing) == {b.address}
-    assert not any(stack for stack in s.stacks.values())
-    assert len(s.ready) + sum(map(len, s.processing.values())) == 20
 
 
-@gen_cluster(Worker=Nanny)
-def test_restart(s, a, b):
-    s.add_client(client='client')
-    s.update_graph(tasks={'x-%d' % i: dumps_task((inc, i)) for i in range(20)},
-                   keys=['x-%d' % i for i in range(20)],
-                   client='client',
-                   dependencies={'x-%d' % i: [] for i in range(20)})
-
-    assert len(s.ready) + sum(map(len, s.processing.values())) == 20
-    assert s.ready
+@gen_cluster(client=True, Worker=Nanny)
+def test_restart(c, s, a, b):
+    futures = c.map(inc, range(20))
+    yield _wait(futures)
 
     yield s.restart()
 
-    for c in [s.stacks, s.processing, s.ncores]:
+    for c in [s.processing, s.ncores, s.occupancy]:
         assert len(c) == 2
 
-    for c in [s.stacks, s.processing]:
+    for c in [s.processing, s.occupancy]:
         assert not any(v for v in c.values())
 
-    assert not s.ready
     assert not s.tasks
     assert not s.dependencies
 
 
-@gen_cluster()
-def test_ready_add_worker(s, a, b):
-    s.add_client(client='client')
-    s.update_graph(tasks={'x-%d' % i: dumps_task((inc, i)) for i in range(20)},
-                   keys=['x-%d' % i for i in range(20)],
-                   client='client',
-                   dependencies={'x-%d' % i: [] for i in range(20)})
-
-    assert all(len(s.processing[w]) == s.ncores[w]
-                for w in s.ncores)
-    assert len(s.ready) + sum(map(len, s.processing.values())) == 20
-
-    w = Worker(s.ip, s.port, ncores=3, ip='127.0.0.1')
-    w.listen(0)
-    s.add_worker(address=w.address, ncores=w.ncores, coerce_address=False)
-
-    assert w.address in s.ncores
-    assert all(len(s.processing[w]) == s.ncores[w]
-                for w in s.ncores)
-    assert len(s.ready) + sum(map(len, s.processing.values())) == 20
-
-    w.scheduler.close_rpc()
-
-
 @gen_cluster()
 def test_broadcast(s, a, b):
     result = yield s.broadcast(msg={'op': 'ping'})
@@ -922,7 +630,6 @@ def test_file_descriptors_dont_leak(s):
 
 @gen_cluster()
 def test_update_graph_culls(s, a, b):
-    s.add_client(client='client')
     s.update_graph(tasks={'x': dumps_task((inc, 1)),
                           'y': dumps_task((inc, 'x')),
                           'z': dumps_task((inc, 2))},
@@ -998,7 +705,7 @@ def test_transition_story(c, s, a, b):
     story = s.transition_story(x.key)
     assert all(line in s.transition_log for line in story)
     assert len(story) < len(s.transition_log)
-    assert all(x.key == line[0] or x.key in line[-1] for line in story)
+    assert all(x.key == line[0] or x.key in line[-2] for line in story)
 
     assert len(s.transition_story(x.key, y.key)) > len(story)
 
@@ -1021,7 +728,7 @@ def test_launch_without_blocked_services():
 def test_scatter_no_workers(c, s):
     with pytest.raises(gen.TimeoutError):
         yield gen.with_timeout(timedelta(seconds=0.1),
-                              s.scatter(data={'x': dumps(1)}, client='alice'))
+                               s.scatter(data={'x': 1}, client='alice'))
 
     w = Worker(s.ip, s.port, ncores=3, ip='127.0.0.1')
     yield [c._scatter(data={'x': 1}),
@@ -1057,3 +764,107 @@ def test_retire_workers(c, s, a, b):
 
     workers = yield s.retire_workers()
     assert not workers
+
+
+@slow
+@pytest.mark.skipif(sys.platform.startswith('win'),
+                    reason="file descriptors not really a thing")
+@gen_cluster(client=True, ncores=[], timeout=240)
+def test_file_descriptors(c, s):
+    psutil = pytest.importorskip('psutil')
+    da = pytest.importorskip('dask.array')
+    proc = psutil.Process()
+    num_fds_1 = proc.num_fds()
+
+    nannies = [Nanny(s.ip, s.port, loop=s.loop) for i in range(20)]
+    yield [n._start() for n in nannies]
+
+    while len(s.ncores) < 20:
+        yield gen.sleep(0.1)
+
+    num_fds_2 = proc.num_fds()
+
+    yield gen.sleep(0.2)
+
+    num_fds_3 = proc.num_fds()
+    assert num_fds_3 == num_fds_2
+
+    x = da.random.normal(10, 1, size=(1000, 1000), chunks=(10, 10))
+    x = c.persist(x)
+    yield _wait(x)
+
+    num_fds_4 = proc.num_fds()
+    assert num_fds_4 < num_fds_3 + 20
+
+    y = c.persist(x + x.T)
+    yield _wait(y)
+
+    num_fds_5 = proc.num_fds()
+    assert num_fds_5 < num_fds_4 + 20
+
+    yield gen.sleep(1)
+
+    num_fds_6 = proc.num_fds()
+    assert num_fds_6 < num_fds_5 + 20
+
+    yield [n._close() for n in nannies]
+
+
+@gen_cluster(client=True)
+def test_learn_occupancy(c, s, a, b):
+    futures = c.map(slowinc, range(1000), delay=0.01)
+    while not any(s.who_has):
+        yield gen.sleep(0.01)
+
+    assert 1 < s.total_occupancy < 40
+    for w in [a, b]:
+        assert 1 < s.occupancy[w.address] < 20
+
+
+@gen_cluster(client=True)
+def test_learn_occupancy_2(c, s, a, b):
+    future = c.map(slowinc, range(1000), delay=0.1)
+    while not any(s.who_has):
+        yield gen.sleep(0.01)
+
+    assert 50 < s.total_occupancy < 200
+
+
+@gen_cluster(client=True, ncores=[('127.0.0.1', 1)] * 30)
+def test_balance_many_workers(c, s, *workers):
+    futures = c.map(slowinc, range(20), delay=0.2)
+    yield _wait(futures)
+    assert set(map(len, s.has_what.values())) == {0, 1}
+
+
+@gen_cluster(client=True, ncores=[('127.0.0.1', 1)] * 30)
+def test_balance_many_workers_2(c, s, *workers):
+    s.extensions['stealing']._pc.callback_time = 100000000
+    futures = c.map(slowinc, range(90), delay=0.2)
+    yield _wait(futures)
+    assert set(map(len, s.has_what.values())) == {3}
+
+
+@gen_cluster(client=True)
+def test_learn_occupancy_multiple_workers(c, s, a, b):
+    x = c.submit(slowinc, 1, delay=0.2, workers=a.address)
+    yield gen.sleep(0.05)
+    futures = c.map(slowinc, range(100), delay=0.2)
+
+    yield _wait(x)
+
+    assert not any(v == 0.5 for vv in s.processing.values() for v in vv)
+    s.validate_state()
+
+
+@gen_cluster(client=True)
+def test_include_communication_in_occupancy(c, s, a, b):
+    s.task_duration['slowadd'] = 0.001
+    x = c.submit(mul, b'0', int(BANDWIDTH), workers=a.address)
+    y = c.submit(mul, b'1', int(BANDWIDTH * 1.5), workers=b.address)
+
+    z = c.submit(slowadd, x, y, delay=1)
+    while z.key not in s.rprocessing:
+        yield gen.sleep(0.01)
+
+    assert s.processing[b.address][z.key] > 1
