@@ -3,16 +3,20 @@ from __future__ import print_function, division, absolute_import
 from concurrent.futures import CancelledError
 from datetime import timedelta
 from operator import add
-from time import sleep, time
+import random
+import sys
+from time import sleep
 
 from dask import delayed
 import pytest
-from toolz import concat, sliding_window
+from toolz import concat, sliding_window, first
 
 from distributed import Client, wait, Nanny
+from distributed.config import config
+from distributed.metrics import time
 from distributed.utils import All
 from distributed.utils_test import (gen_cluster, cluster, inc, slowinc, loop,
-        slowadd, slow)
+        slowadd, slow, slowsum)
 from distributed.client import _wait
 from tornado import gen
 
@@ -41,6 +45,8 @@ def test_stress_gc(loop, func, n):
             assert x.result() == n + 2
 
 
+@pytest.mark.skipif(sys.platform.startswith('win'),
+                    reason="test can leave dangling RPC objects")
 @gen_cluster(client=True, ncores=[('127.0.0.1', 1)] * 4, timeout=None)
 def test_cancel_stress(c, s, *workers):
     da = pytest.importorskip('dask.array')
@@ -100,6 +106,7 @@ def test_stress_creation_and_deletion(c, s):
 @gen_cluster(ncores=[('127.0.0.1', 1)] * 10, client=True, timeout=60)
 def test_stress_scatter_death(c, s, *workers):
     import random
+    s.allowed_failures = 1000
     np = pytest.importorskip('numpy')
     L = yield c._scatter([np.random.random(10000) for i in range(len(workers))])
     yield c._replicate(L, n=2)
@@ -124,15 +131,29 @@ def test_stress_scatter_death(c, s, *workers):
             s.validate_state()
         except Exception as c:
             logger.exception(c)
-            import pdb; pdb.set_trace()
+            if config.get('log-on-err'):
+                import pdb; pdb.set_trace()
+            else:
+                raise
         w = random.choice(alive)
         yield w._close()
         alive.remove(w)
 
     try:
-        yield gen.with_timeout(timedelta(seconds=10), c._gather(futures))
+        yield gen.with_timeout(timedelta(seconds=25), c._gather(futures))
     except gen.TimeoutError:
-        import pdb; pdb.set_trace()
+        ws = {w.address: w for w in workers if w.status != 'closed'}
+        print(s.processing)
+        print(ws)
+        print(futures)
+        try:
+            worker = [w for w in ws.values() if w.waiting_for_data][0]
+        except:
+            pass
+        if config.get('log-on-err'):
+            import pdb; pdb.set_trace()
+        else:
+            raise
     except CancelledError:
         pass
 
@@ -141,18 +162,56 @@ def vsum(*args):
     return sum(args)
 
 
+@pytest.mark.avoid_travis
 @slow
 @gen_cluster(client=True, ncores=[('127.0.0.1', 1)] * 80, timeout=1000)
 def test_stress_communication(c, s, *workers):
     s.validate = False # very slow otherwise
     da = pytest.importorskip('dask.array')
+    # Test consumes many file descriptors and can hang if the limit is too low
+    resource = pytest.importorskip('resource')
+    try:
+        soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
+        lim = 8192
+        if soft < lim:
+            resource.setrlimit(resource.RLIMIT_NOFILE, (lim, max(hard, lim)))
+    except Exception as e:
+        pytest.skip("file descriptor limit too low and couldn't be increased :"
+                    + str(e))
 
     n = 40
     xs = [da.random.random((100, 100), chunks=(5, 5)) for i in range(n)]
     ys = [x + x.T for x in xs]
-    z = da.atop(vsum, 'ij', *concat(zip(ys, ['ij'] * n)))
+    z = da.atop(vsum, 'ij', *concat(zip(ys, ['ij'] * n)), dtype='float64')
 
     future = c.compute(z.sum())
 
     result = yield future._result()
     assert isinstance(result, float)
+
+
+@pytest.mark.skip
+@gen_cluster(client=True, ncores=[('127.0.0.1', 1)] * 10, timeout=60)
+def test_stress_steal(c, s, *workers):
+    s.validate = False
+    for w in workers:
+        w.validate = False
+
+    dinc = delayed(slowinc)
+    L = [delayed(slowinc)(i, delay=0.005) for i in range(100)]
+    for i in range(5):
+        L = [delayed(slowsum)(part, delay=0.005)
+             for part in sliding_window(5, L)]
+
+    total = delayed(sum)(L)
+    future = c.compute(total)
+
+    while future.status != 'finished':
+        yield gen.sleep(0.1)
+        for i in range(3):
+            a = random.choice(workers)
+            b = random.choice(workers)
+            if a is not b:
+                s.work_steal(a.address, b.address, 0.5)
+        if not s.processing:
+            break
