@@ -1,21 +1,25 @@
 from __future__ import print_function, division, absolute_import
 
 from concurrent.futures import CancelledError
+from datetime import timedelta
 from operator import add
 import os
-from time import time, sleep
+from time import sleep
 
-from dask import delayed
 import pytest
 from toolz import partition_all
 from tornado import gen
 
+from dask import delayed
+from distributed import Client, Nanny, wait
 from distributed.compatibility import PY3
 from distributed.client import _wait
+from distributed.core import coerce_to_address
+from distributed.metrics import time
+from distributed.nanny import isalive
 from distributed.utils import sync, ignoring
 from distributed.utils_test import (gen_cluster, cluster, inc, loop, slow, div,
         slowinc, slowadd)
-from distributed import Client, Nanny, wait
 
 
 def test_submit_after_failed_worker(loop):
@@ -48,10 +52,14 @@ def test_gather_then_submit_after_failed_workers(loop):
             total = c.submit(sum, L)
             wait([total])
 
-            (_, port) = first(c.scheduler.who_has[total.key])
+            addr = c.who_has()[total.key][0]
+            _, port = coerce_to_address(addr, out=tuple)
             for d in [x, y, z]:
                 if d['port'] == port:
                     d['proc'].terminate()
+                    break
+            else:
+                assert 0, "Could not find worker"
 
             result = c.gather([total])
             assert result == [sum(map(inc, range(20)))]
@@ -65,7 +73,7 @@ def test_failed_worker_without_warning(c, s, a, b):
     original_process = a.process
     a.process.terminate()
     start = time()
-    while a.process is original_process and a.process.poll() is not None:
+    while a.process is original_process and not isalive(a.process):
         yield gen.sleep(0.01)
         assert time() - start < 10
 
@@ -106,8 +114,9 @@ def test_restart(c, s, a, b):
     f = yield c._restart()
     assert f is c
 
-    assert len(s.stacks) == 2
     assert len(s.processing) == 2
+    assert len(s.occupancy) == 2
+    assert not any(s.occupancy.values())
 
     assert not s.who_has
 
@@ -130,7 +139,8 @@ def test_restart_cleared(c, s, a, b):
     yield c._restart()
 
     for coll in [s.tasks, s.dependencies, s.dependents, s.waiting,
-            s.waiting_data, s.who_has, s.restrictions, s.loose_restrictions,
+            s.waiting_data, s.who_has, s.host_restrictions,
+            s.worker_restrictions, s.loose_restrictions,
             s.released, s.priority, s.exceptions, s.who_wants,
             s.exceptions_blame]:
         assert not coll
@@ -236,6 +246,13 @@ def test_multiple_clients_restart(s, a, b):
     yield e2._shutdown(fast=True)
 
 
+@gen_cluster(Worker=Nanny)
+def test_restart_scheduler(s, a, b):
+    import gc; gc.collect()
+    yield s.restart()
+    assert len(s.ncores) == 2
+
+
 @gen_cluster(Worker=Nanny, client=True)
 def test_forgotten_futures_dont_clean_up_new_futures(c, s, a, b):
     x = c.submit(inc, 1)
@@ -247,7 +264,7 @@ def test_forgotten_futures_dont_clean_up_new_futures(c, s, a, b):
     yield y._result()
 
 
-@gen_cluster(client=True, timeout=60)
+@gen_cluster(client=True, timeout=60, active_rpc_timeout=10)
 def test_broken_worker_during_computation(c, s, a, b):
     n = Nanny(s.ip, s.port, ncores=2, loop=s.loop)
     n.start(0)
