@@ -2,9 +2,12 @@ from __future__ import print_function, division, absolute_import
 
 from collections import Iterable
 from contextlib import contextmanager
+import inspect
 import logging
+import multiprocessing
 import os
 import re
+import shutil
 import socket
 import six
 import sys
@@ -28,6 +31,14 @@ from .config import config
 logger = logging.getLogger(__name__)
 
 
+if PY3 and not sys.platform.startswith('win'):
+    mp_context = multiprocessing.get_context('forkserver')
+    # Makes the test suite much faster
+    mp_context.set_forkserver_preload(['distributed'])
+else:
+    mp_context = multiprocessing
+
+
 def funcname(func):
     """Get the name of a function."""
     while hasattr(func, 'func'):
@@ -38,6 +49,24 @@ def funcname(func):
         return str(func)
 
 
+def has_arg(func, argname):
+    """
+    Whether the function takes an argument with the given name.
+    """
+    while True:
+        try:
+            if 'dask_worker' in inspect.getargspec(func).args:
+                return True
+        except TypeError:
+            break
+        try:
+            # For Tornado coroutines and other decorated functions
+            func = func.__wrapped__
+        except AttributeError:
+            break
+    return False
+
+
 def get_fileno_limit():
     """
     Get the maximum number of open files per process.
@@ -62,7 +91,7 @@ def get_ip(host='8.8.8.8', port=80):
 def ignoring(*exceptions):
     try:
         yield
-    except exceptions:
+    except exceptions as e:
         pass
 
 
@@ -170,6 +199,7 @@ def is_kernel():
 
 hex_pattern = re.compile('[a-f]+')
 
+
 def key_split(s):
     """
     >>> key_split('x')
@@ -221,6 +251,14 @@ def key_split(s):
         return 'Other'
 
 
+try:
+    from functools import lru_cache
+except ImportError:
+    pass
+else:
+    key_split = lru_cache(100000)(key_split)
+
+
 @contextmanager
 def log_errors(pdb=False):
     try:
@@ -313,25 +351,35 @@ def iterator_to_queue(seq, maxsize=0):
 
 
 def tokey(o):
-    """ Convert an object to a bytestring, using str
+    """ Convert an object to a string.
 
     Examples
     --------
 
     >>> tokey(b'x')
-    b'x'
+    'x'
     >>> tokey('x')
     'x'
     >>> tokey(1)
     '1'
     """
     t = type(o)
-    if t is str or t is bytes:
+    if t is str:
         return o
+    elif t is bytes:
+        return o.decode('latin1')
     else:
         return str(o)
 
 
+def validate_key(k):
+    """Validate a key as received on a stream.
+    """
+    if type(k) is not str:
+        raise TypeError("Unexpected key type %s (value: %r)"
+                        % (type(k), k))
+
+
 def _maybe_complex(task):
     """ Possibly contains a nested task """
     return (istask(task) or
@@ -357,26 +405,6 @@ def str_graph(dsk, extra_values=()):
     return {tokey(k): convert(v) for k, v in dsk.items()}
 
 
-import logging
-from .compatibility import logging_names
-logging.basicConfig(format='%(name)s - %(levelname)s - %(message)s',
-                    level=logging.INFO)
-
-for name, level in config.get('logging', {}).items():
-    LEVEL = logging_names[level.upper()]
-    logging.getLogger(name).setLevel(LEVEL)
-
-# http://stackoverflow.com/questions/21234772/python-tornado-disable-logging-to-stderr
-stream = logging.StreamHandler(sys.stderr)
-stream.setLevel(logging.CRITICAL)
-logging.getLogger('tornado').addHandler(stream)
-logging.getLogger('tornado').propagate = False
-
-
-from contextlib import contextmanager
-import shutil
-
-
 def seek_delimiter(file, delimiter, blocksize):
     """ Seek current file to next byte after a delimiter bytestring
 
